#include "process.h"
#include "gui.h"

// --- PROCESS GLOBAL VARIABLES --- //

//struct cdatastruct {	// Classes Data
//	unsigned int Nc;	// number of classes 
// 	unsigned int c;		// current (pointed class) class
//	unsigned int fdim;	// dimension of feature vector
//	mmatrix *fv;		// feature data values
//} cpar = {0, 0, 3, NULL};

cdatastruct cpar = {0, 0, 3, NULL};

CvImage //img,		// Original image (is now defined as static within main)
		imgIRB,		// YCrCb img (Intensity - Red Chrominance - Blue Chrominance)
		imgSegMask,	// Segmentation mask (image)
		imgSeg,		// Segmentation image (original image and segmented pixels)
		imgG;		// Gray scale image

CvScalar ColorTransform(CvScalar color_in, int code)
{
	CvImage pixelcolor( cvSize(1,1), IPL_DEPTH_8U, 3);  // create a singe pixel image for color conversion
	//pixelcolor.create( img.size(), IPL_DEPTH_8U, 3 ); // another method to create an image - from scratch
	cvSet1D( pixelcolor, 0, color_in ); //CV_RGB(color)
	cvCvtColor( pixelcolor, pixelcolor, code); //CV_BGR2YCrCb
	CvScalar color_out = cvGet1D( pixelcolor, 0 );
	return color_out;
}

// Computes feature vector.
// Here the feature vector is composed of Y, Cr and Cb values
// (intensity, red chrominance and blue chrominance components).
mvector* ComputeFeature(CvImage *img )
{
	//A MTT 20090802 CvScalar Iv; // store intensities
	unsigned char* ptr_u; // B
	long offset; // C
	mvector *v = mvectoralloc( (*img).channels() );

	// Clear reference feature vector
	for(int chn = 0; chn < (*img).channels(); chn++)
	{
		(*v).val[chn] = 0.0;
	}

	// Calculate reference feature vector
	ptr_u = (unsigned char *) (img->data()) + img->step()*rect.ymin + rect.xmin*img->channels(); //added MTT20090208 //C
	offset = img->step()-(rect.xmax-rect.xmin)*img->channels(); // C
	for(int row = rect.ymin; row < rect.ymax; row++ )
	{
		//ptr_u = (unsigned char *) (img->data()) + img->step()*row + rect.xmin*img->channels(); //added MTT20090208 // B
		for(int col = rect.xmin; col < rect.xmax; col++)
		{
			//MTT20090208 Iv = cvGet2D( (*img), row, col ); // A
			for(int chn = 0; chn < (*img).channels(); chn++)
			{

				//MTT20090208 (*v).val[chn] = (*v).val[chn]+Iv.val[chn]; // A
				//(*v).val[chn]+=CV_IMAGE_ELEM( (IplImage *) (*img), unsigned char, row, col*img->channels()+chn ); // D
				//(*v).val[chn]+=((unsigned char*)( img->data() + img->step()*row ))[col*img->channels() + chn]; // E
				(*v).val[chn]+= *ptr_u++; // B or C
			}
		}
		ptr_u+=offset; // C
	}
	
	if (bboxarea(rect)>0)
	{
		for(int chn = 0; chn < (*img).channels(); chn++)
		{
			(*v).val[chn] = (*v).val[chn]/bboxarea(rect);
			printf( "v[%d] = %f\n", chn, (*v).val[chn] );
		}
	}
	else
	{
		printf(	"Bounding box area is zero!\n"
				"Please use the mouse right-button to select a region\n"
				"before computing a feature vector\n" );
	}
	
	return v;
}

// Segment the image.
// The segmentation is implemented in terms of a simple nearest-neighbor type classifier.
void ComputeSegmentation(CvImage* img, cdatastruct *cpar)
{
	CvScalar Iv; // store intensities

	imgSeg = img->clone(); // copy img to imgSeg

	cvSetZero(imgSegMask);

	// Perform the color segmentation
	for(int row = 0; row < img->height(); row++ )
	{
		for(int col = 0; col < img->width(); col++)
		{
			Iv = cvGet2D( imgIRB, row, col );
			//printf("%d - %d\n", idx0, (int) Iv.val[1]);

			for(unsigned int j=0; j<(*cpar).Nc; j++) // for each class
			{
				if (   ( ( (*cpar).fv[0].val[1][j]-cr_thresh < Iv.val[1]) && (Iv.val[1] < (*cpar).fv[0].val[1][j]+cr_thresh) )
					&& ( ( (*cpar).fv[0].val[2][j]-cr_thresh < Iv.val[2]) && (Iv.val[2] < (*cpar).fv[0].val[2][j]+cr_thresh) )
					)
				{
					//cvSet2D( imgSeg, row, col, ColorTransform(color_table[j], CV_BGR2YCrCb));
					cvSet2D( imgSeg, row, col, color_table[j]);
					//printf("(%d, %d)\n", row, col);
					cvSet2D( imgSegMask, row, col, color_table[j] );
				}
			}
		}
	}
}

void AddClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		(*cpar).fv = mmatrixalloc((*cpar).fdim,1);
		for(unsigned int i=0;i<(*cpar).fdim;i++)
			(*cpar).fv[0].val[i][0] = 0.0;
		(*cpar).Nc = 1;
		(*cpar).c = 0;
	}
	else
	{
		mmatrix *fv_aux; // aux storage
		fv_aux = mmatrixalloc((*cpar).fdim,(*cpar).Nc);
		for(unsigned int j=0;j<(*cpar).Nc;j++) // move existing elements to auxiliary storage
		{
			for(unsigned int i=0;i<(*cpar).fdim;i++)
				(*fv_aux).val[i][j] = (*cpar).fv[0].val[i][j];
		}
		mmatrixfree( (*cpar).fv );
		(*cpar).fv = mmatrixalloc((*cpar).fdim,(*cpar).Nc+1);
		for(unsigned int j=0;j<(*cpar).Nc;j++) // retrieve existing elements from auxiliary storage
		{
			for(unsigned int i=0;i<(*cpar).fdim;i++)
                (*cpar).fv[0].val[i][j] = (*fv_aux).val[i][j];
		}
		for(unsigned int i=0;i<(*cpar).fdim;i++)
			(*cpar).fv[0].val[i][(*cpar).Nc] = 0.0;
		(*cpar).Nc++;
		(*cpar).c++;
		mmatrixfree( fv_aux );
	}
}

void RemoveClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot remove current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		mmatrix *fv_aux; // aux storage
		fv_aux = mmatrixalloc((*cpar).fdim,(*cpar).Nc);
		for(unsigned int j=0;j<(*cpar).Nc;j++) // move existing elements to auxiliary storage
		{
			for(unsigned int i=0;i<(*cpar).fdim;i++)
				(*fv_aux).val[i][j] = (*cpar).fv[0].val[i][j];
		}
		mmatrixfree( (*cpar).fv );
		(*cpar).fv = mmatrixalloc((*cpar).fdim,(*cpar).Nc-1);
		for(unsigned int j=0, k=0;j<(*cpar).Nc;j++) // retrieve existing elements from auxiliary storage
		{
			if (j != (*cpar).c)
			{
				for(unsigned int i=0;i<(*cpar).fdim;i++)
					(*cpar).fv[0].val[i][k] = (*fv_aux).val[i][j];
				k++;
			}
		}
		(*cpar).Nc--;
		if ((*cpar).Nc == 0)
			(*cpar).c=0;
		else if ( (*cpar).c == (*cpar).Nc )
			(*cpar).c--;

		mmatrixfree(fv_aux);
	}
}

void PrintClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot print existing classes.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		printf("Total number of classes Nc = %d\n", (*cpar).Nc);
		printf("Current pointed class c = %d\n", (*cpar).c+1);
		for(unsigned int j = 0; j<(*cpar).Nc; j++)
		{
			printf("cpar.fv->val[.][%d]\n",j);
			for(unsigned int i = 0; i<(*cpar).fdim; i++)
				printf("\t.val[%d] = %f\n", i, (*cpar).fv->val[i][j]);
			printf("\n");
		}
	}
}

void UpdateClass(cdatastruct *cpar)
{
	mvector *v = NULL;

	if ((*cpar).Nc == 0)
	{
		printf("Cannot update current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		if ( (*cpar).c >= (*cpar).Nc ) (*cpar).c = (*cpar).Nc-1;
		if ( (*cpar).c < 0 ) (*cpar).c = 0;
		v = ComputeFeature( &imgIRB );
		for(unsigned int i=0;i<(*cpar).fdim;i++)
			(*cpar).fv[0].val[i][(*cpar).c] = (*v).val[i];
	}

	mvectorfree(v);
}

void RandomlyFillClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot randomly fill current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		for(unsigned int i=0; i<(*cpar).fdim; i++)
			(*cpar).fv[0].val[i][(*cpar).c] = (double) rand()/(double) RAND_MAX;
	}
}

void BlankClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot blank current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		for(unsigned int i=0;i<(*cpar).fdim;i++)
			(*cpar).fv[0].val[i][(*cpar).c] = 0.0;
	}
}

void DetectClasses(CvImage* img, cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot perform the segmentation of the color classes.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		ComputeSegmentation( img, cpar );
	}
}

void IncrementClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot increment current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		(*cpar).c++;
		if ((*cpar).c > (*cpar).Nc-1)
			(*cpar).c = 0;
		printf("Current class is %d.\n", (*cpar).c+1);
	}
}

void DecrementClass(cdatastruct *cpar)
{
	if ((*cpar).Nc == 0)
	{
		printf("Cannot decrement current class.\nPlease first add a class using 'a'.\n");
	}
	else
	{
		if ((*cpar).c == 0)
			(*cpar).c = (*cpar).Nc-1;
		else
			(*cpar).c--;
		printf("Current class is %d.\n", (*cpar).c+1);
	}
}

void WriteCPD(cdatastruct *cpar, char* filename) // Write Class Parameters Data
{   
	FILE *output = fopen( filename, "w" );
	if (output != NULL)
	{
		fprintf(output, "Nc %u FvDim %u\n", (*cpar).Nc, (*cpar).fdim);
		for(unsigned int j=0; j<(*cpar).Nc; j++)
		{
			for(unsigned int i=0; i<(*cpar).fdim; i++)
			{
				fprintf(output, "%f", (*cpar).fv[0].val[i][j] );
				if ( i == (*cpar).fdim-1 ) fprintf(output, "\n");
				else fprintf(output," ");
			}
		}
		fclose(output);
	}
}

void ReadCPD(cdatastruct *cpar, char* filename) // Read Class Parameters Data
{
	FILE *input = fopen( filename, "r" );
	unsigned int Nc, fdim;

	if ( (*cpar).fv != NULL)
		mmatrixfree( (*cpar).fv );

	if (input != NULL)
	{
		fscanf(input, "Nc %u FvDim %u\n", &Nc, &fdim);  //&(*cpar).Nc, &(*cpar).fdim);
		(*cpar).fv = mmatrixalloc( fdim, Nc );
		(*cpar).Nc = Nc;
		(*cpar).fdim = fdim;
		(*cpar).c = 0;
		printf("Nc %u FvDim %u\n", (*cpar).Nc, (*cpar).fdim);
		for(unsigned int j=0; j<(*cpar).Nc; j++)
		{
			for(unsigned int i=0;i<(*cpar).fdim; i++)
			{
				if ( i == (*cpar).fdim-1 ) fscanf(input, "%lf\n", &(*cpar).fv[0].val[i][j] ); // Use %lf to read doubles (not %f)!!!
				else fscanf(input, "%lf", &(*cpar).fv[0].val[i][j]); // Use %lf to read doubles (not %f)!!!
				if ( i == (*cpar).fdim-1 ) printf("%f\n", (*cpar).fv[0].val[i][j] );
				else printf("%f ", (*cpar).fv[0].val[i][j] );
			}
		}
        fclose(input);
	}
}

int ProcessColorSegmentation(CvImage *img, cmdlinestruct cmdlineparam)
{
	imgIRB = img->clone();
	cvCvtColor( *img, imgIRB, CV_BGR2YCrCb);
	DetectClasses( img, &cpar );

	if (cmdlineparam.output_flag)
	{
		char straux[512];
		sprintf(straux,"%s_mask.jpg", cmdlineparam.basefilename );
		cvSaveImage(straux, imgSegMask);
	}

	if (cmdlineparam.verbose_flag)
	{
		cvCvtColor( imgSegMask, imgG,  CV_BGR2GRAY );
		cvCanny( imgG, imgG, 1.1, 5.9, 2*edge_thresh+3 );
	}

	return 0;
}

int ProcessColorConversion(CvImage *img)
{
	imgIRB = img->clone();
	cvCvtColor( *img, imgIRB, CV_BGR2YCrCb);			
	cvCvtColor( *img, imgG,  CV_BGR2GRAY );
	cvCanny( imgG, imgG, 1.1, 5.9, 2*edge_thresh+3 );

	return 0;
}