#include <stdio.h>
//#include <gsl/gsl_math.h>
//#include "cv.h"       // use " " instead of < > to get list of class/struct members
//#include "highgui.h"  // use " " instead of < > to get list of class/struct members
#include "test.h"
//#include "mathlib.h"
//#include "dataconversion.h"
//#include "util.h"
#include "datastructures.h"
#include "process.h"

int test_timing(void)
{
	double freq, t0, elapsed;

	freq = timer_hr_freq();
	t0 = timer_hr_count();
	delay(0.05);
	elapsed = timer_hr_elapsed(t0);

	printf("Frequency of the high-resolution performance counter [Counts per sec]: %.0f\n", freq);
	printf("Frequency of the high-resolution performance counter [MHZ]: %f\n", freq/1e6);
	printf("Resolution of the high-resolution performance counter [us]: %f\n", 1e6/freq);
	printf("Current value of the high-resolution performance counter [counts]: %.0f\n", t0);
	printf("Elapsed time of the high-resolution performance counter [ms]: %f\n", elapsed);
	
	return 0;
} // test_timing()

int test_vector_matrix(void)
{
	printf("Testing vector structure...\n");
	mvector *v;
	v=mvectoralloc(10);
	mvectorfilltest(v);
	mvectorprint(v);
	mvectorfree(v);

	printf("Testing matrix structure...\n");
	mmatrix *m;
	m=mmatrixalloc(3,5);
	mmatrixfilltest(m);
	mmatrixprint(m);
	mmatrixfree(m);
	
	return 0;
}

int test_structure(void)
{
	printf("Testing cdatastruct (Class data structure)...\n");
	cdatastruct cd = {5, 0, 3, NULL};
	cd.fv = mmatrixalloc(3,5);
	mmatrixfilltest( cd.fv );
	cd.fv->val[1][2]=pi_const;
	mmatrixprint( cd.fv );
	mmatrixfree( cd.fv );

	return 0;
}

int test_functions(void)
{	
	cdatastruct cd = {0, 0, 3, NULL};

	printf("Testing functions to create and manipulate the cdatastruct...\n");
	printf("1. Add three classes and fill them with random values as they are added...\n");
	cd.fdim = 2;
	AddClass( &cd );
	RandomlyFillClass( &cd );
	AddClass( &cd );
	RandomlyFillClass( &cd );
	AddClass( &cd );
	RandomlyFillClass( &cd );
	mmatrixprint( cd.fv );

	printf("2. Decrement current class index twice and fill current class with random values...\n");
	DecrementClass( &cd );
	DecrementClass( &cd );
	RandomlyFillClass( &cd );	
	mmatrixprint( cd.fv );

	printf("3. Increment current class index once and remove current class...\n");
	IncrementClass( &cd );
	RemoveClass( &cd );
	mmatrixprint(cd.fv);

	printf("4. Test class data displaying and reading/writing from/to file...\n");
	PrintClass( &cd );
	WriteCPD( &cd, "test_cpd.txt" );
	BlankClass( &cd );  // Current class is blanked to show that saved values are restored correctly
	PrintClass( &cd );
	ReadCPD( &cd, "test_cpd.txt" );
	PrintClass( &cd );

	printf("5. Test conversion from static array to dynamic array using pointers...\n");
	// First static matrices are declared and initialized inline in two different ways.
	double mp1[][3] = {{ 1.0, 2.0, 3.0},
  	                   { 4.0, 5.0, 6.0}};
	// The following does not work, even if it is conceptually equivalent to the previous declaration.
	//double v1[3] = { 1.0, 2.0, 3.0 },
	//	     v2[3] = { 4.0, 5.0, 6.0 };
	//double mp2[][3] = { {v1}, {v2} };
	// Use instead a typedef
	typedef double v3[3];
	v3 mp2[] = {{ 1.0, 2.0, 3.0 }, { 4.0, 5.0, 6.0 }}; 

	// Static matrices can be converted into dynamic matrices by assigning the address of thstatic matrix
	// to an appropriate pointer.  This can be achieved by any of the three following approaches, 
	// which rely on i) creating a pointer to the matrix rows, ii) create a vector of pointers to rows.
	// Note that pointer-to-pointers need  some memory to be allocated somehow to store the vector
	// of pointers.  Attemting something simple as:
	//    double m[2][3];
	//    double **m_ptr;
	//    m_ptr = &m;
	// does not work.
	//
	// Approach 1: Short, but harder to read/understand
	///*	
	double *dptr[2];  // Array of 2 double-pointers.  This also can be written as "double* dptr[2]".
					  // Note that the data type of the dptr[2] array is 'double pointer'.
					  // This also can be achieved by:
					  //	typedef double* dpt;  // declares a data type 'double *' called dpt
					  //	dpt dptr[2];          // creates an array of 2 'dpt' elements
	double (*p3)[3];  // Pointer to array of 3 doubles.
					  // This is a shortcut to:
					  //    typedef double array[3];
					  //    array *p3;
					  // -----------------------------------
					  // Note the difference:
	                  //   'int *ptr[10]' means that ptr is the name of an array of
	                  //                  ten pointers to type double.
	                  //   'int (*ptr)[10]' means that ptr is the name of a pointer
	                  //                  to an array of 10 int.
					  // Therefore:
					  //   typedef double array[3];
					  //   array *a;
					  //   typedef array* ap[2];
					  //   ap mx;
				      // is defining mx as a "matrix" of two rows of pointers 'ap' that contain the pointers
					  // to the three element arrays 'array'.
					  // -----------------------------------
					  // !!! From the above, passing the address of a fixed array 'double mp[][]' to an array
	                  //     in pointer format 'double **' can be achieved by any of the following two 
					  //     approaches.
	p3 = &(mp1[0]);           // point to the first row of matrix mp1
	dptr[0] = (double *) p3;
	p3 = &(mp1[1]);
	dptr[1] = (double *) p3;  // point to the first row of matrix mp1
	double **mm = (double **) &dptr;
	//*/

	// Approach 2: Technically correct and easier to read, but longer
	/*
	double **mm;
	typedef double array[3];
	array *a;
	typedef array* ap[2];
	ap mx;
	a = &(mp1[0]);
	mx[0] = a;
	a = &(mp1[1]);
	mx[1] = a;
	mm = (double **) mx;
	*/

	//Approach 3: Using malloc, in a way equivalent to that for creating matrices; see mmatrixalloc() above.
	/*
	double **mm;
	double **mma = (double **) malloc(2*sizeof(double *));
	double *mr  = (double *) malloc(3*sizeof(double));
	mr = (double *) &mp2[0];
	mma[0] = mr;
	mr = (double *) &mp2[1];
	mma[1] = mr;
	mm = mma;
	*/

	mmatrix m = {2, 3, NULL};
	m.val = mm;	
	mmatrixprint( &m );

	mmatrix* matrix_aux = cd.fv; // temporarily store the pointer in cd to the feature vectors
	cd.Nc=3;
	cd.fdim=2;
	cd.fv = &m;                  // assing new data to the feature vectors
	PrintClass( &cd );
	cd.fv = matrix_aux;          // restore pointer to original data, so that it can be de-allocated.
	                             // If the original pointer is not restored, the matrixfree()
	                             // function will attempt to de-allocate a static variable, 
	                             // producing an error.  On the other hand, de-allocating the data
								 // storage reserved by AddClass() above would not be released.
	mmatrixfree( cd.fv );

	return 0;
}

int test_read_sup( int argc, char** argv )
{
	char *supfile,          // Start-up parameters (SUP) file
		 *buffer;           // SUP file buffer	
	char val[300];          // Auxiliary string storage for parameters read from the
	char straux[256];

	int Nc=0;               // Number of classes 
	
	int dLimit = 3;	// FilterSize/4 - Distance from the center mask pixel
								// to the mask boudaries.
	int xmin, ymin, xmax, ymax;	// Auxiliary variables to store the region.
								// Not really essential, they help to make the
								// code "more" readable in some parts.
	int xsize = 800;        // Image width
	int ysize = 640;		// Image height

	//--- Checking CPU Type ---
	DisplayByteOrder();
	
	// --- Read Switch Arguments ---
	supfile = read_switch(&argc, argv, "-s", 1, "defsup.txt");


	// --- Load Start-up Parameters ---
	printf("--- Reading Start-Up Parameters ---\n");
	buffer=loadenv(supfile);
	//printf("File contents are:\n%s\n",buffer);
	
	// --- Get number of classes ---
	Nc=atoi(get_env_val(&val[0], buffer, "Nc=", "0"));
	printf("Nc=%i\n",Nc);
	if (Nc<=0)
	{
	 fprintf(stderr,"Invalid number of classes.  Check 'Nc=' in file %s!\n",supfile);
	 exit(1);
	}

	printf("--- Reading class-field region coordinates ---\n");
	classdataArray cset;
	cset.Nc = Nc;

	// Allocate memory for class set of fields.
	cset.cdataptr = (classdata *) malloc( Nc*sizeof(classdata) );
    
    // Read number of fields in class k.
    for(int c=0; c<Nc; c++)
    {
       sprintf(straux,"nc%if=",c+1);
       cset.cdataptr[c].Nfield=atoi(get_env_val(&val[0], buffer,straux,"-1"));
       if (cset.cdataptr[c].Nfield<0)
       {
          fprintf(stderr,"Invalid class field.  Check '%s' in file %s!\n",straux,supfile);
          exit(1);
       }
       printf("nc%if: %i\n",c+1,cset.cdataptr[c].Nfield);

	   cset.cdataptr[c].field=(region *) malloc(cset.cdataptr[c].Nfield*sizeof(region));
	   cset.cdataptr[c].fieldNelems=(int *) malloc(cset.cdataptr[c].Nfield*sizeof(int));
       // Read coordinates for field f of class c and calculate the field area.
	   for(int f=0; f<cset.cdataptr[c].Nfield; f++)
       {
          sprintf(straux,"c%if%i=",c+1,f+1);
          sprintf(straux,"%s",get_env_val(&val[0], buffer, straux, 
					  "[0 0]x[0 0]"));
		  regionParse(&cset.cdataptr[c].field[f],straux);
		  xmin=cset.cdataptr[c].field[f].xmin;
		  ymin=cset.cdataptr[c].field[f].ymin;
		  xmax=cset.cdataptr[c].field[f].xmax;
		  ymax=cset.cdataptr[c].field[f].ymax;
          printf("c%if%i = [%i %i]x[%i %i].\n", c+1, f+1, xmin, ymin, xmax, ymax);
          // Check field coordinates are within the processable image area
          if ((xmin-2*dLimit<0) || (ymin-2*dLimit<0) ||
              (xmax+2*dLimit>xsize-1) || (ymax+2*dLimit>ysize-1))
          {
             fprintf(stderr, "main(): Class %i - Field %i: Ensure training field coordinates are at least 2*dLimit pixel from image boundary!\n",c+1, f+1);
             exit(1);
          }
          // Check validity of field coordinates
          if (((xmax-xmin+1)<1) || ((ymax-ymin+1)<1))
          {
             fprintf(stderr, "main(): Class %i - Field %i: Invalid coordinates! Check xmin<xmax and ymin<ymax.\n", c+1, f+1);
             exit(1);
          }
          // Calculate number of elements per class fields
		  cset.cdataptr[c].fieldNelems[f] = (xmax-xmin+1)*(ymax-ymin+1);
		  printf("\tfieldNelems: %i\n", cset.cdataptr[c].fieldNelems[f]);
       } // end f (fields loop)
    } // end c (classes loop)
	
	// Deallocate memory
	for(int c=0; c<Nc; c++)
	{
		free( cset.cdataptr[c].field );
		free( cset.cdataptr[c].fieldNelems );
	}
	free( cset.cdataptr );
	
	free(buffer);

	return 0;
}

int test_ground( int argc, char** argv )
{
	test_timing();
	test_read_sup( argc, argv );

	//test_vector_matrix();
	//test_structure();
	//test_functions();

	//cvWaitKey();

	return 0;
}