#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>	//MTT20090208   required for ptrdiff_t data type
#include <time.h>
#include "util.h"
#include "windows.h"

//struct cmdlinestruct {		// Stores command line options set by the user
//	char basefilename[256];		// Frame base filename
//	char fileextension[250];	// Frame file extension
//	char filenameformat[200];	// "%s%0<n>d.%s" (filename)(zero-padding number)(extension)
//	long int iv;				// Frame initial value
//	long int fv;				// Frame final value
//	double fps;					// Video writer parameters
//	char filename_out[512];		// Output filename
//	bool output_flag;			// 1 = save results to disk
//	bool display_flag;			// 1 = display results
//	bool verbose_flag;			// 1 = print process status in the command line interface
//	bool pause_flag;			// 1 = pause and wait for user to confirm actions
//} cmdlineparam;

cmdlinestruct cmdlineparam;

// --------------------------------------------------------------------------
// Very basic utility functions.
//
bool atob(char* str) // ascii to bool
{
	int number = atoi(str);

	if (number>0)
		return (bool) 1;
	else
		return (bool) 0;
}

void fatalError (char *message)
{
	fprintf( stderr, "%s\n", message );
	exit(1);
} // fatalError()


// --------------------------------------------------------------------------
// Functions for checking the byte order (little endian or big endian)
// and swaping the byte order.  These can useful for reading raw data
// indpendently from the given CPU's byte order.
//
/*********************** (FORMERLY) STATIC FUNCTIONS ***********************/

/************************************************
 * Function:    byteswap_SR_IR                  *
 *   Author:    Dave Hascher (Veridian Inc.)    *
 *     Date:    06/05/97                        *
 *    Email:    dhascher@dytn.veridian.com      *
 ************************************************
 * 'SR' --> Sun 32-bit float value              *
 * 'IR' --> PC-Intel 32-bit float value         *
 ************************************************/

float byteswap_SR_IR(unsigned char *pointer)
{
  float *temp;
  unsigned char iarray[4], *charptr;

  iarray[0] = *(pointer + 3);
  iarray[1] = *(pointer + 2);
  iarray[2] = *(pointer + 1);
  iarray[3] = *(pointer );
  charptr = iarray;
  temp    = (float *) charptr;
  return *(temp);
}


/************************************************
 * Function:    byteswap_SUS_IUS                *
 *   Author:    John Querns (Veridian Inc.)     *
 *     Date:    06/05/97                        *
 *    Email:    jquerns@dytn.veridian.com       *
 ************************************************
 * 'SUS' --> Sun 16-bit uns short value         *
 * 'IUS' --> PC-Intel 16-bit uns short value    *
 ************************************************/

unsigned short byteswap_SUS_IUS(unsigned char *pointer)
{
  unsigned short *temp;
  unsigned char iarray[2], *charptr;

  iarray[0] = *(pointer + 1);
  iarray[1] = *(pointer );
  charptr = iarray;
  temp    = (unsigned short *) charptr;
  return *(temp);
}


/**********************************
 *   checkByteOrder()             *
 **********************************
 * Taken from:                    *
 *                                *
 *   Encyclopedia of Graphic File *
 *   Formats, Murray & Van Ryper, *
 *   O'Reilly & Associates, 1994, *
 *   pp. 114-115.                 *
 *                                *
 * Desc: Checks byte-order of CPU.*
 **********************************/

int CheckByteOrder(void)
{
  short   w = 0x0001;
  char   *b = (char *) &w;

  return(b[0] ? LSB_FIRST : MSB_FIRST);
}

void byteswap(unsigned char *ptr_in, int BytesPerWord)
{
  unsigned char temp;

  for(int i=0;i <BytesPerWord/2;i++)
  {
     temp = ptr_in[i];
     ptr_in[i] = ptr_in[BytesPerWord - i - 1];
     ptr_in[BytesPerWord - i -1] = temp;
  }
}

void DisplayByteOrder(void)
{
	int byteorder;
	byteorder=CheckByteOrder();
	switch(byteorder)
	{
	case LSB_FIRST:
		printf("Little-Endian CPU detected.\n");
		break;
	case MSB_FIRST:
		printf("Big-Endian CPU detected.\n");
		break;
	}
}
/***************************************************************************/

// --------------------------------------------------------------------------
// Basic file openning and closing functions.
//
FILE *openFile (char *filename, char *options) 
{
 char errorMsg[80];
 FILE *filePtr;

        filePtr = (FILE *) fopen (filename, options);
        if (filePtr == (FILE *) NULL) {
		sprintf (errorMsg, "openFile: fopen() failed to open file %s\n",
				filename);
		fatalError (errorMsg);
		}
	return (filePtr);
} // openFile()

void closeFile (FILE *FP)
{
  if (fclose(FP) != (int) 0)
     fatalError ("closeFile:  fclose() returned a non-zero value.\n");
} // closeFile()


// --------------------------------------------------------------------------
// Functions for handling the program options given as command line arguments.
//
/*************************************************************************
 *                                                                       *
 *                       Routine: read_switch()                          *
 *                        Author: Otto Milvang, U of Oslo, Norway        *
 *                          Date: 1990...                                *
 *                                                                       *
 *           This routine is an exact copy of the routine called         *
 *           "read_switch()" found in the B-LAB XITE source routine      *
 *           "readswitch.c", which was written by Otto Milvang, U.       *
 *           of Oslo, Image Processing lab, Dept of Informatics,         *
 *           Copyright 1990. Hence, I am including the following         *
 *           copyright notice...                                         *
 * ________________________________________________________________      *
 *                                                                       *
 *       Copyright 1990, Blab, UiO                                       *
 *       Image processing lab, Department of Informatics                 *
 *       University of Oslo                                              *
 *       E-mail: blab@ifi.uio.no                                         *
 * ________________________________________________________________      *
 *                                                                       *
 *                                                                       *
 * Permission to use, copy, modify and distribute this software and its  *
 * documentation for any purpose and without fee is hereby granted,      *
 * provided that this copyright notice appear in all copies and that     *
 * both that copyright notice and this permission notice appear in       *
 * supporting documentation and that the name of B-lab, Department of    *
 * Informatics or University of Oslo not be used in advertising or       *
 * publicity pertaining to distribution of the software without specific,*
 * written prior permission.                                             *
 *                                                                       *
 * B-LAB DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING*
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT    *
 * SHALL B-LAB BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL      *
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR *
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER        *
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR      *
 * PERFORMANCE OF THIS SOFTWARE.                                         *
 *************************************************************************
 *                                                                       *
 *  This portion of the documentation is essentially my own.  I include  *
 *  it to better help you to understand how to use the subroutine.       *
 *                                                                       *
 *   John Querns, Veda Inc, (937) 255-1116, Ext 2818                     *
 *   email: jquerns@mbvlab.wpafb.af.mil                                  *
 *                                                                       *
 *  read_switch()                                                        *
 *                                                                       *
 *  DESCRIPTION:  This is a nifty subroutine for extracting arguments    *
 *                using switches. I just extracted it from the main C    *
 *                routine to make it more "standalone" for my purposes.  *
 *                                                                       *
 *      FORMAT:   char *read_switch( int* argc,                          * 
 *                                 char** argv,                          *
 *                                 char* name,                           *
 *                                 int args,                             *
 *                                 char* defreturn );                    *
 *                                                                       *
 *                'read_switch' reads a switch from the command line.    *
 *                routine searches for the switch 'name'.  If 'name' is  *
 *                found, and 'args' is zero (i.e. args following switch),*
 *                a pointer to the switch is returned.  If 'name' is     *
 *                and 'args' is non-zero, a string containing the switch *
 *                argument(s) is returned.  If 'name' does not occur in  *
 *                the command line, 'defreturn' is returned.             *
 *                                                                       *
 *                The "nifty" thing about this subroutine is that it only*
 *                looks at the "variable" arguments.  As it finds them,  *
 *                it decrements "argc" till only the REQUIRED arguments  *
 *                are left.  Also, it keeps you from having to fool      *
 *                around trying to parse out the command line.           *
 *                                                                       *
 *    EXAMPLES:                                                          *
 *                                                                       *
 *       #include <math.h>                                               *
 *                                                                       *
 *        char *res, *filnam;                                            *
 *        double d;                                                      *
 *        int i, j, l;                                                   *
 *                                                                       *
 *        d = atof(read_switch(&argc, argv, "-scale", 1, "0.5"));        *
 *        sscanf(read_switch(&argc, argv, "-offset", 2, "128 128"),      *
 *                          "%d%d", &i, &j);                             *
 *        l = read_switch(&argc, argv, "-log", 0, NULL) != NULL;         *
 *        res = read_switch(&argc, argv, "-yes", 0, "-no");              *
 *        filnam = argv[1];                                              *
 *                                                                       *
 *        As you can see, since "read_switch" returns a string, you have *
 *        to use other routines (i.e. atof, atoi, sscanf, etc.) to get   *
 *        the arg values in the proper form....                          *
 *                                                                       *
 *************************************************************************/

//#include <stdio.h>
//#include <stdlib.h>

char *read_switch(int *argc, char **argv,
                  char *name, int args, char *defreturn)
{
 int i, argn;
  char *ret;
  for(i=1; i<*argc; i++)
    {
      if (strcmp(name, argv[i]) == 0)
      {
        argn = i+1;
        if (i + args >= *argc)
          {
            fprintf(stderr,"read_switch: Missing argument after \"%s\"\n",
                    argv[i]);
            exit(2);
          }
        ret = argv[args ? i+1 : i];
        argn = i + 1 + args;
        while(argn < *argc)
            argv[i++] = argv[argn++];
        *argc = *argc - argn + i;
        if (args-- <=1) return(ret);
        while(args--) ret[strlen(ret)] = ' ';
        return(ret);
      }
    }
  return(defreturn);
} // read_switch()


// Initializes the cmdlinestruct structure
int initcmdlinestruct(cmdlinestruct* cmdlineparam)
{
	sprintf((*cmdlineparam).basefilename,"%s","input");		// Frame base filename
	sprintf((*cmdlineparam).fileextension,"%s","jpg");		// Frame file extension
	sprintf((*cmdlineparam).filenameformat,"%%s%%0%dd.%%s",3);	// Builds file suffix format string "%s%0<n>d.%s"
																// where <n> is the number for zero padding.
	(*cmdlineparam).iv=-1;									// Frame initial value
	(*cmdlineparam).fv=-1;									// Frame final value
	(*cmdlineparam).fps=1e6;								// Video writer parameters
	sprintf((*cmdlineparam).filename_out,"%s","output.jpg");	// Output filename
	(*cmdlineparam).output_flag=1;
	(*cmdlineparam).display_flag=1;
	(*cmdlineparam).verbose_flag=1;
	(*cmdlineparam).pause_flag=1;
	return 0;
} // initcmdlinestruct()

// Prints help message
int printhelp(void)
{
	printf(
			"ptz                   -> captures video from camera if connected,\n"
			"                         otherwise loads image 'input.jpg'.\n"
			"ptz [camera_number]   -> process video captured with the n-th webcam\n"
			"                         connected to the PC.\n"
			"ptz <video.avi>       -> process AVI video file 'video.avi'\n"
			"ptz <base filename> [options] [output filename=base_filename.jpg]\n"
			"                      -> process input file.\n"
			"For example:\n"
			"ptz f -e jpg -sq 0 10 -> loads files f000.jpg, f001.jpg ... f010.jpg\n"
			"\n"
			"Description:\n"
			"\tOpens a file, a file sequence, a video file or video stream from a camera,\n"
			"\tdisplays the images and applies a simple color segmentation process in the\n"
			"\n"
			"Options:\n"
			"\t-h   Prints this help.\n"
			"\t-nv  Turns verbose off.\n"
			"\t-np  Do not wait for user to press ESC when done.\n"
			"\t-nd  Do not display images.\n"
			"\t-no  Do not save the output.\n"
			"\t-e   Filename extension [=jpg]\n"
			"\t-sq initial_value final_value\n"
			"\t    Tells ptz to process a sequence of images\n"
			"\tstarting at filename<initial_value>.ext and\n"
			"\tending in filename<initial_value>.ext\n"
			"\tBy default the sequence range is assumed to be between 0 and 10\n"
			"\twith zero-padding = 3, i.e. 000, 001, ..., 010.\n"
			"\t-zp num   Pad filename suffix with num zeros.\n"
			"\t-fps num  Frames per seconds [=max_fps].\n"
			"\n"
			"Author: Miguel Torres Torriti (c) 2008.01.15, v. 1.\n"
		);
	return -2;
} // printhelp()

// Parses the cmdlinestruct structure
int parsecmdline(int argc, char** argv, cmdlinestruct* cmdlineparam)
{
	// Parse command line options
	if (read_switch(&argc, argv, "-h", 0, NULL)) 
		return printhelp(); // print help

	// atob returns 0 if read_switch finds the string, because
	// read_switch returns a string without numbers
	cmdlineparam->verbose_flag = atob(read_switch(&argc, argv, "-nv", 0, "1"));
	cmdlineparam->pause_flag = atob(read_switch(&argc, argv, "-np", 0, "1"));
	cmdlineparam->display_flag = atob(read_switch(&argc, argv, "-nd", 0, "1"));
	cmdlineparam->output_flag = atob(read_switch(&argc, argv, "-no", 0, "1"));

	// Get file extension
	sprintf((*cmdlineparam).fileextension, "%s", read_switch(&argc, argv, "-e", 1, "jpg"));
	
	// Get sequence limits
	sscanf(read_switch(&argc, argv, "-sq", 2, "-1 -1"),
		"%d%d", &((*cmdlineparam).iv), &((*cmdlineparam).fv));

	// Get zero-padding
	sprintf((*cmdlineparam).filenameformat, "%%s%%0%dd.%%s", 
		atoi(read_switch(&argc, argv, "-zp", 1, "3")));

	// Get frames per second
	cmdlineparam->fps = atof(read_switch(&argc, argv, "-fps", 1, "1e6"));

	// If there are extra options...
	if (argc>1)
	{
		for (int i=1; i<argc; i++)
		{	
			if (strcspn( argv[i], "-" )==0) // warn about invalid options
			{
				if ((*cmdlineparam).verbose_flag) 
					printf("Invalid option detected: %s\n", argv[i]);
			}
			else if (i==1)
			{	
				// Generate the output filename tokenizing the input filename
				sprintf((*cmdlineparam).basefilename, "%s", argv[i]);
				if ((*cmdlineparam).verbose_flag) printf("New input base filename: %s\n", (*cmdlineparam).basefilename);
			}
			else
			{
				sprintf((*cmdlineparam).filename_out, "%s", argv[i]);
				if ((*cmdlineparam).verbose_flag) printf("New output filename: %s\n", (*cmdlineparam).filename_out);
			}
		}
	}

	// Print flags and options
	if ((*cmdlineparam).verbose_flag)
	{
		printf("Verbose is on.\n");
		if ((*cmdlineparam).pause_flag) printf("Pausing is on.\n");
		if ((*cmdlineparam).display_flag) printf("Displaying is on.\n");
		if ((*cmdlineparam).output_flag) printf("Output will be saved.\n");
		
		printf("Base filename: %s\n", cmdlineparam->basefilename);
		printf("File extension: %s\n", cmdlineparam->fileextension);
		printf("Output filename: %s\n", cmdlineparam->filename_out);
		printf("File zero-padding format: %s\n", cmdlineparam->filenameformat);
		printf("Sequence v0: %d, v1: %d\n",(*cmdlineparam).iv,(*cmdlineparam).fv);
		printf("FPS set to: %f\n", cmdlineparam->fps);
	}
	
	return 0;
} // parsecmdline()

// Prints the welcome message
int printwelcome(void)
{
	printf ("ptz: Project template illustrating a deterministic color-based segmentation process %s (%d.%d.%d)\n",
    CV_VERSION, CV_MAJOR_VERSION, CV_MINOR_VERSION, CV_SUBMINOR_VERSION);

	printf( "Hot keys: \n"
		    "\tESC - quit the program\n"
			"\tMouse-L - show position and intensity values of selected pixel\n"
			"\tMouse-R - select region\n"
			"\tMouse-R+CTRL - draw on buffer\n"
			"\tp - process current image using loaded or precomputed data\n"
			"\ta - add class\n"
			"\tr - remove current class\n"
			"\tl - list existing classes\n"
			"\tu - calculate feature vector for current region and update the class parameters\n"
			"\tb - blank current class data\n"
			"\tv - fill current class data with random values\n"
			"\t< - decrease current class\n"
			"\t> - increase current class\n"
			"\th - toggle process holding (temporarily disables/enables processing)\n"
			"\tg - get class data from cpd.txt\n"
			"\ts - save class data to cpd.txt\n"
			"\tt - test basic functions\n"
			"\t\n");
	return 0;
} // printwelcome()

// Builds the filename for sequences of file
int makefilename(cmdlinestruct* cmdlineparam, char* filename, int frame_number)
{
	if ((*cmdlineparam).iv<0)
		sprintf(filename,"%s.%s", (*cmdlineparam).basefilename, (*cmdlineparam).fileextension);
	else
		sprintf(filename,(*cmdlineparam).filenameformat,(*cmdlineparam).basefilename,frame_number,(*cmdlineparam).fileextension);

	return 0;
} // makefilename()


// --------------------------------------------------------------------------
// Functions for loading and processing the program environment file.
//
/*
 * char *loadenv()
 * Loads the file with start-up parameters or environment variables
 * storing its contents in a buffer allocated according to the size of
 * the file.
 *
 * RETURNS
 * A pointer to char pointing the location of the buffer.
 *
 * Note: The buffer must be deallocated using free, before the end
 *       of the program.
 *
 * Author: Miguel Torres Torriti (c) 2001 v. 1.0
 */
char *loadenv(char *filename)
{
 FILE *f;
 char *buffer, errorMsg[80];
 int c;
 unsigned long fsize=0;

 if ( (f=fopen(filename, "rt")) == NULL)
 {
    sprintf(errorMsg, "Unable to open environment file: %s.\n", filename);
    fatalError(errorMsg);
 }

 /* --- Find the size of the file --- */
 // This also can be done using the POSIX but not ANSI 'fstat' function.
 while( (c=fgetc(f)) != EOF) // Note: 'c' MUST be int, since EOF is not char.
    fsize++;
 rewind(f); //use also fseek or fsetpos
 // The equivalent approach cannot be used since 'fseek' has odd behaviours
 // under MS-DOS, in this case a fsize of around 18, 19 bytes additional
 // to the real size if returned.
 //fseek(f,0,SEEK_END);
 //fsize=ftell(f);
 //rewind(f);
 printf("%i bytes in start-up file %s read successfully.\n", fsize, filename);

 /* Allocate buffer memory to store the file*/
 buffer=(char *) malloc((fsize+1)*sizeof(char));
 fread(buffer, sizeof(char), fsize, f);
 sprintf(&buffer[fsize],"\0");

 fclose(f);
 return buffer;
} // loadenv()

/*
 * char *get_env_val()
 *
 * RETURNS
 * The a string with the value of the program environment or start-up parameter
 * 'var_name'.
 *
 * The function searches for 'var_name' in 'buffer', pruning comments which
 * start with a '#' sign that can be located anywhere in the line.  If the
 * '#' is at the beginning of the line, the whole line will be disregarded,
 * in the same way the // comment marks work in C++.
 * Since the returned value is a pointer to a string, the output might
 * require further processing with functions like atoi, sscanf to convert
 * the values to an appropriate form.
 * This routine simplifies the parsing of the environment parameters.
 * It was created to avoid long command-line arguments and was inspired
 * on the read_switch() routine by Otto Milvang included below as reference.  
 *
 * Usage: get_env_val(buffer, "parameter_name=", "default_value");
 *
 * Notes: At the present time, this routine looks for whatever exactly
 *        replaces "parameter_name=", which means that the parameter file
 *        must contain parameters in the form
 *                     parameter1=value1
 *                     Param XYZ=abc
 *        * If the parameters file contains lines of the form "paramN =10",
 *          and one is looking for "paramN=", no matching result will be
 *          found.  This caveat certainly can be improved... with some
 *          additional effort.
 *        * Conflictive parameter names...
 *          Do not name parameters in the following manner:
 *            xyz_abc=1
 *            abc=123
 *          because searching for "abc=123", will yield the first
 *          occurrence of "abc=" which is in "xyz_abc=1", hence
 *          the result will be 1 instead of 123.
 *          ! BE CAREFUL as naming parameters in such a way is not 
 *            that unusual, e.g.
 *            set_time = true
 *            time = 102030
 *
 * Author: Miguel Torres Torriti (c) 2001 v. 1.0
 */
char *get_env_val(char *val, char *buffer, char *var_name, char *defreturn)
{
 char line[300], *posinline, *posbuffer, *ret;
 unsigned long bytesread = 0, bytesline = 0, buffersize = (unsigned long) strlen(buffer);
 //MTT20090208   int dpos;
 ptrdiff_t dpos;

 posbuffer = buffer;

 while(bytesread<buffersize)
 {// It is necessary to clear the "line" buffer, otherwise sscanf will
  // not clear it when reading an empty line (a line with no characters
  // before a carriage return).
  sprintf(line,"");
  sscanf(posbuffer,"%[^\n]\n",line);
  strcat(line,"\n");
  bytesline = (unsigned long) strlen(line);
  bytesread+=bytesline;
  if (bytesread<buffersize)
  {  // Point to new position in buffer according to bytes read in last line.
     //MTT20090208   posbuffer = (char *) ((unsigned long) posbuffer + bytesline);
	posbuffer+=bytesline;
  }

  // Look for comments and truncate the line after the comment mark '#'.
  if ( (posinline=strchr(line,'#')) != NULL)
  {
      //MTT20090208   dpos= (int) ((unsigned long) posinline - (unsigned long) line);
      dpos= posinline - line; 
	  line[dpos]='\n';
        line[dpos+1]='\0';
  }

  // Find var_name in the line.
  if ( (ret=strstr(line, var_name)) != NULL)
  {
    //MTT20090208   ret=(char *) ((unsigned long) ret + strlen(var_name));
    ret+=strlen(var_name);
	// The following 2 lines were replaced by the next 2, due to problems
	// under Unix, which were not occurring under MS-DOS/Win.
	// Also the input var: 'val' was added to this function. 
	//sscanf(ret,"%[^\n]",ret);
	//return(ret); // Return the reminder of the line after the occurrence
                   // of var_name.
	sscanf(ret,"%[^\n]",val);
	
	// Remove trailing spaces
	while((val[strlen(val)-1]==' ')||(val[strlen(val)-1]=='\t'))
	{
	  val[strlen(val)-1]='\0';
	}
	return(val);
  }

 } // end while */

 return(defreturn);
} // get_env_val()

/*
 * void regionParse()
 *
 * RETURNS
 * The (xmin, ymin) x (xmax, ymax) coordinates in a structure of type 'region'
 * extracted from a string as returned by get_env_val().
 *
 * The string must match the format "[xmin ymin]x[xmax ymax]".
 *
 * Usage: regionParse(region, "[4 23]x[25 77]);
 *
 * Notes: The first coordinate pair MUST be the upper left coordinate
 *        in the standar image coordinates, while the second coordinate
 *        pair must be the lower right coordinate, i.e. for any region
 *        [px py]x[qx qy], the following must be satisfied:
 *                      px <= qx and py <= qy
 *
 * Author: Miguel Torres Torriti (c) 2001 v. 1.0
 */
void regionParse(region *r, char *str)
{
 sscanf(str,"[%i %i]x[%i %i]",&r->xmin,&r->ymin,&r->xmax,&r->ymax);
}


// --------------------------------------------------------------------------
// Some other functions for string parsing and manipulation.
//
/*
 *  Parse a string and return the tokens
 *
 *	Uses whitespace separation ONLY
 *
 *  Author: Ian Burns, Guy Smith.  MeasTex, v. 1.1, 1997.05.27 
 */
int parseString (char *string, char **tokenPtrs)
{
 int  numTokens;
 char *tempPtr;

	numTokens = 0;
	tempPtr = (char *) strtok (string, " \t");
	if (!tempPtr) {
		return (0);
		}
	while ((tokenPtrs[numTokens] = tempPtr) != NULL) {
		tempPtr = (char *) strtok ((char *) NULL, " \t");
		numTokens++;
		}
	return (numTokens);
} /* parseString() */


/*
 *  Parse a string and return the tokens
 *
 *	User specified delimiters
 *
 *  Author: Ian Burns, Guy Smith.  MeasTex, v. 1.1, 1997.05.27
 */
int parseStringDelim (char *string, char **tokenPtrs, char *delims)
{
 int  numTokens;
 char *tempPtr;

	numTokens = 0;
	tempPtr = (char *) strtok (string, delims);
	if (!tempPtr) {
		return (0);
		}
	while ((tokenPtrs[numTokens] = tempPtr) != NULL) {
		tempPtr = (char *) strtok ((char *) NULL, delims);
		numTokens++;
		}
	return (numTokens);
} /* parseStringDelim() */


/*
 *      Convert string to upper case
 *
 *  Author: Ian Burns, Guy Smith.  MeasTex, v. 1.1, 1997.05.27
 */
char *stringUpper (char *string)
{
 int i;

	if (!string)
		return ((char *) NULL);
	for (i = 0; string[i]; i++) {
		if (string [i] >= 'a' && string [i] <= 'z')
			string[i] -= 32;
	}
	return (string);
} /* stringUpper() */


// --------------------------------------------------------------------------
// Timing functions.
//
void delay(double secs)
{
  clock_t t0, t1;
  
  t0=clock();
  t1=t0;
  while ((double) (t1-t0) / (double) CLOCKS_PER_SEC < secs)
  {
    t1=clock();
  }
} // delay()

// Returns current value of the high resolution performance counter
double timer_hr_count(void)
{
	LARGE_INTEGER lpPerformanceCount;

	QueryPerformanceCounter(&lpPerformanceCount);
	
	return (double) lpPerformanceCount.QuadPart;
}

// Returns high resolution performance counter frequency in counts per second
double timer_hr_freq(void)
{
	LARGE_INTEGER lpFrequency;

	QueryPerformanceFrequency(&lpFrequency);
	
	return (double) lpFrequency.QuadPart;
}

// Returns elapsed time in seconds.  The input must be the initial count value
// of the high resolution performance counter as returned by function timer_hr_count().
double timer_hr_elapsed(double initial_count)
{
	LARGE_INTEGER lpPerformanceCount;
	
	QueryPerformanceCounter(&lpPerformanceCount);
	
	return ((double) lpPerformanceCount.QuadPart - initial_count)/timer_hr_freq();
}

